"""
This module contains functions for simulating the evolution of genomes on a tree.
"""

import matplotlib.pyplot as plt
import numpy as np
import cgt
from cgt import *
from networkx.drawing.nx_agraph import graphviz_layout
from networkx import json_graph
import networkx as nx


def newick_to_tree(newick_string):
    json_tree = cgt.parsers.newick_to_json(
        newick_string, generate_names=True, lengthstring="weight"
    )
    tree = json_graph.tree_graph(json_tree)
    return tree


def evolve_on_tree(tree, framework, model, root="random", at_least_one_change=False, exactly_n_changes=None):
    # The tree should be a networkx DiGraph
    # Choose a genome at the root of the tree
    if root == "random":
        root = framework.random_genome()
    else:
        root = framework.identity_genome()

    # Get the root and set the genome at the root
    root_id = [n for n, d in tree.in_degree() if d == 0][0]
    tree.nodes[root_id]["genome"] = root

    # Traverse the tree
    root = [n for n, d in tree.in_degree() if d == 0][0]
    for successor, node in dict(nx.bfs_predecessors(tree, root)).items():
        # Get the genome at the parent node
        parent = tree.nodes[node]["genome"]
        # Get the branch length of the in-edge
        branch_length = tree.nodes[successor]["weight"]
        tree[node][successor]["weight"] = branch_length
        tree[node][successor]["len"] = branch_length
        # Decide how many rearrangements to apply
        if exactly_n_changes is not None:
            num_rearrangements = exactly_n_changes
        else:
            num_rearrangements = np.random.poisson(branch_length)
        if at_least_one_change:
            num_rearrangements = max(num_rearrangements, 1)
        # Apply them
        tree.nodes[successor]["genome"] = parent
        if num_rearrangements == 0:
            tree.nodes[successor]["label"] = str(
                framework.collect_genome_terms(parent)
            ).replace(" ", "")
            tree.nodes[successor]["genome"] = parent
        for r in range(num_rearrangements):
            # Draw a rearrangement
            rearrangement = draw_rearrangement(model)
            # Apply the rearrangement
            formal_sum = tree.nodes[successor]["genome"] * rearrangement
            selected_genome_instance = cgt.simulations.draw_genome(formal_sum)
            canonical_instance = framework.canonical_instance(selected_genome_instance)
            genome = framework.genome(
                framework.cycles(
                    framework.canonical_instance(selected_genome_instance)
                ),
                format=cgt.FORMAT.formal_sum,
            )
            # Update the genome at the successor node
            tree.nodes[successor]["genome"] = genome
            tree.nodes[successor]["label"] = str(framework.collect_genome_terms(genome))

    return tree


def grow_tree(framework, model, num_taxa):
    model_element = model.s_element()
    # Start with the identity genome
    genome = framework.genome_group().identity()
    print(framework.draw_instance(genome, shortened=True))
    # Instantiate the tree
    G = nx.DiGraph()
    G.add_node(0, label=str(genome), genome=genome)  # initialize root
    for t in range(0, num_taxa + 1, 2):
        # Select a genome from the tree
        node = np.random.choice([node for node in G.nodes() if G.out_degree(node) == 0])
        parent = G.nodes[node]["genome"]

        # Draw a branch length From a multinomial distribution
        # Draw the number of events from a poisson distribution
        # Choose the rearangements from the model and apply them

        # Grow two new leaves
        for t_1 in range(2):
            # Draw a rearrangement
            rearrangement = draw_rearrangement(model)
            # Apply the rearrangement
            genome = parent * rearrangement
            # Draw a branch length
            branch_length = draw_branch_length()
            # Add the new genome to the tree
            G.add_node(t + t_1 + 1, label=str(genome), genome=genome)
            G.add_edge(node, t + t_1 + 1, weight=round(branch_length, 1))

    # Return the tree
    return G


def draw_tree(tree):
    # Draw the tree
    plt.figure(figsize=(9, 5))
    G = tree
    pos = graphviz_layout(G, prog="dot")  # , args='-Granksep=0.1 -Gnodesep=0.1')
    labels = nx.get_node_attributes(G, "label")
    edge_labels = nx.get_edge_attributes(G, "weight")
    nx.draw(G, pos=pos, node_size=0)
    nx.draw_networkx_labels(
        G,
        pos,
        labels=labels,
        bbox={"facecolor": "white", "edgecolor": "white"},
        font_size=10,
    )
    nx.draw_networkx_edge_labels(
        G, pos=pos, rotate=False, edge_labels=edge_labels, font_size=10
    )
    plt.show()


def draw_branch_length():
    return np.random.exponential(1)


def draw_rearrangement(model):
    model_element = model.s_element()
    rearrangements, probs = tuple(zip(*[list(x)[0] for x in (model_element).terms()]))
    return np.random.choice(rearrangements, p=probs)


def draw_genome(formal_sum):
    genomes, probs = tuple(zip(*list(formal_sum)))
    return np.random.choice(genomes, p=probs)


def set_node(node, framework, genome):
    node["genome"] = genome
    node["label"] = str(framework.canonical_instance(genome))



"""
"""

from enum import Enum, auto

class MODEL(Enum):
    # Inversions
    all_inversions         = "all inversions"
    one_region_inversions  = "inversions of a single region"
    two_region_inversions  = "inversions of two adjacent regions"
    # Transpositions
    all_transpositions  = "all transpositions"
    ## "Swap" interpretation
    two_region_transpositions    = "Two region adjacent transpositions, with or without inversion"
    two_region_transpositions_without_inversions = "Two region adjacent transpositions, without inversion"
    two_region_revrevs = "Take two adjacent regions and invert both of them"
    ## "Move" interpretation
    one_region_moves  = "Move a single region, with or without inversion"
    one_region_moves_without_inversions  = "Move a single region, without inversion"
    # Other transpositions
    three_region_transpositions = "transpositions involving a segment of three regions, with or without inversion"
    # Deprecated options
    two_region_adjacent_transpositions = "transpositions of two adjacent regions"
    # Special models
    all_inversions_larger_less_likely = "all inversions, with larger inversions less likely"

class TYPE(Enum):
    reg_to_signed_pos = auto()
    pos_to_signed_reg = auto()
    signed_reg_to_pos = auto()
    signed_pos_to_reg = auto()

class FORMAT(Enum):
    formal_sum     = auto()
    equiv_classes  = auto()
    dictionary     = auto()
    only_reps      = auto()

class SET(Enum):
    signed_cycles 	= auto()
    unsigned_cycles = auto()
    one_row			= auto()
    wreath			= auto()
    wreath_s2		= auto()
    
class SYMMETRY(Enum):
    circular = dihedral = D_n              = auto()
    linear = S_2 = C_2 = flip = reflection = auto()

class CLASSES(Enum):
    conjugacy_classes = auto()
    cosets            = auto()
    double_cosets     = auto()

class DISPLAY(Enum):
    arrows  = auto()
    one_row = auto()
    cycles  = auto()

class ALGEBRA(Enum):
    group        = auto()
    genome       = auto()
    genome_class = auto()

class IRREP_TYPE(Enum):
    specht = speck = auto()
    orthogonal     = auto()
    seminormal     = auto()

class DISTANCE(Enum):
    min = minimum = minimum_distance = auto()
    min_weighted = minimum_weighted = auto()
    MFPT = mean_first_passage_time = auto()
    MLE = maximum_likelihood_distance = maximum_likelihood_estimate = auto()
    discrete_MFPT = DMFPT = discrete_mean_first_passage_time = auto()

class DATA(Enum):
    eig_data = 'eigen_data'
    eigval_lists = 'eigenvalue_lists'
    eigval_sets = 'eigenvalue_sets'
    eigvec_lists = 'eigenvector_lists'
    eigvec_mat_inv = "eigenvector_matrix_inverses"
    projections = "projections"
    eigval_sets_old = 'eig_lists'
    irreps_z = 'irreps_of_z'
    irreps_zs = 'irreps_of_zs'
    reg_rep_of_zs = 'reg_rep_of_zs'
    partial_traces = 'partial_traces'
    irreps_z_np = 'irreps_of_z_np'



from .. import example, distances, position_paradigm, Model
from .. import MODEL
import numpy as np
import pytest

def test_likelihood_function():
    for n in (3,4,5,6):
        framework = position_paradigm.GenomeFramework(n)
        converge_point = framework.symmetry_group().order()/framework.genome_group().order()
        model = Model.named_model_with_relative_probs(framework, {
            MODEL.all_inversions: 1
        })
        for _ in range(10):
            instance = framework.cycles(framework.canonical_instance(framework.random_instance()))
            identity_instance = framework.identity_instance()
            L = distances.likelihood_function(framework, model, instance)
            print(L(0))
            print(instance)
            if instance == identity_instance:
                assert np.isclose(L(0), 1)
            else:
                assert np.isclose(L(0), 0)
            assert np.isclose(float(L(500)), float(converge_point))


def test_small_inversions_min_dist():
    for n in (3,4):
        framework, model = example(n)
        min_distances_graph_based = distances.min_distance(framework, model)
        min_distance_irrep_based = distances.min_distance_using_irreps(framework, model)
        for instance, distance in min_distances_graph_based.items():
            assert distance == min_distance_irrep_based[instance]



"""

"""
from copy import deepcopy
from sage.all_cmdline import SymmetricGroup, floor
from . import structures
from .enums import CLASSES, SYMMETRY


def inversion(framework, about_position, length):
    """
    Return an instance of the inversion that inverts a segment in a genome.

    Args:
        framework (PositionParadigmFramework): The framework to create the inversion in
        about_position (int): The position to invert about
        length (int): The length of the segment to invert

    Returns:
        Permutation: An instance representing the inversion that inverts the segment
    """
    if not framework.oriented or framework.symmetry != SYMMETRY.circular:
        raise NotImplementedError(f"not yet implemented for {str(framework)}")
    n = framework.n
    r = about_position
    string_rep = ""
    if length % 2 == 0:
        k = int(length / 2)
        for i in range(1, k + 1):
            string_rep += (
                f"({(r-(i-1)) % n}, -{(r+i) % n})(-{(r-(i-1)) % n}, {(r+i) % n})"
            )
    else:
        k = int((length - 1) / 2)
        string_rep += f"({about_position},{-about_position})"
        for i in range(1, k + 1):
            string_rep += f"({(r-i) % n},-{(r+i) % n})(-{(r-i) % n},{(r+i) % n})"
    string_rep = string_rep.replace("0", str(n))
    return framework.cycles(string_rep)


signed_inversion = inversion  # old name


def all_inversion_instances(framework, length=None):
    """
    Return all instances that represent an inversion of the given length

    Args:
        framework (PositionParadigmFramework): The framework to create the inversion in
        length (int, optional): The length of the inversion. Defaults to None (all inversions).

    Returns:
        set: The set of all instances that represent an inversion of the given length
    """
    if not framework.oriented or framework.symmetry != SYMMETRY.circular:
        raise NotImplementedError(f"not yet implemented for {str(framework)}")

    n = framework.n
    rearrangements = []

    for r in range(1, n + 1):
        for l in (length,) if length is not None else range(1, n + 1):
            rearrangements.append(inversion(framework, r, l))

    return rearrangements


def transposition(framework, sec_1, sec_2, inv_1=False, inv_2=False, revrev=False):
    """
    Return an instance of the transposition that swaps two segments in a genome.

    Args:
        framework (PositionParadigmFramework): The framework to create the transposition in
        sec_1 (tuple): The first section to transpose (tuple of start (included) and end (excluded) positions)
        sec_2 (tuple): The second section to transpose (tuple of start (included) and end (excluded) positions)
        inv_1 (bool, optional): Whether to also invert the first section. Defaults to False.
        inv_2 (bool, optional): Whether to also invert the second section. Defaults to False.

    Returns:
        Permutation: An instance representing the transposition that swaps the two segments
    """
    if not framework.oriented or framework.symmetry != SYMMETRY.circular:
        raise NotImplementedError(f"not yet implemented for {str(framework)}")
    if all((inv_1, inv_2)) and not revrev:
        raise ValueError(
            "for circular genomes, inverting both segments is not allowed. Explicitly allow revrevs by setting allow_revrev=True"
        )
    if revrev:
        if not all((inv_1, inv_2)):
            raise ValueError("revrevs are only allowed if both segments are inverted")
        else:
            inv_1 = inv_2 = False
    n = framework.n

    if sec_1[1] != sec_2[0]:
        raise ValueError("sections must be adjacent")
    full_inversion = (
        signed_inversion(
            framework,
            segment_midpoint(n, sec_1[0], sec_2[1]),
            segment_length(n, sec_1[0], sec_2[1]),
        )
        if not revrev
        else framework.cycles("()")
    )
    left_inversion = (
        signed_inversion(
            framework, segment_midpoint(n, *sec_1), segment_length(n, *sec_1)
        )
        if not inv_1
        else framework.cycles("()")
    )
    right_inversion = (
        signed_inversion(
            framework, segment_midpoint(n, *sec_2), segment_length(n, *sec_2)
        )
        if not inv_2
        else framework.cycles("()")
    )
    return left_inversion * right_inversion * full_inversion


def all_transposition_instances(
    framework,
    scope_limit=None,
    single_segment_limit=None,
    with_inversion=True,
    include_revrevs=False,
    only_revrevs=False,
    canonical_reps_only=False,
):
    """
    Return all instances that represent a transposition of the given length

    Args:
        framework (PositionParadigmFramework): The framework to create the transposition in
    Returns:
        set: The set of all instances that represent a transposition
    """
    if not framework.oriented or framework.symmetry != SYMMETRY.circular:
        raise NotImplementedError(f"not yet implemented for {str(framework)}")

    n = framework.n
    rearrangements = set()

    if scope_limit is None:
        scope_limit = n - 1

    for full_length in range(2, scope_limit + 1):  # Scope of the rearrangement
        for start in range(1, n + 1):  # Start of first segment
            if single_segment_limit is not None:
                first_segment_lengths = sum(
                    list(range(1, single_segment_limit + 1)),
                    list(range(full_length - single_segment_limit, full_length)),
                )
            else:
                first_segment_lengths = range(1, full_length)

            for l1 in first_segment_lengths:  # Length of first segment
                l2 = full_length - l1  # length of second segment
                middle = (start + l1 - 1) % n + 1
                end = (middle + l2 - 1) % n + 1
                s, m, e = start, middle, end
                possible_transpositions = (
                    {transposition(framework, (s, m), (m, e))}
                    if not only_revrevs
                    else set()
                )
                if with_inversion and not only_revrevs:
                    possible_transpositions |= {
                        transposition(framework, (s, m), (m, e), inv_1=True),
                        transposition(framework, (s, m), (m, e), inv_2=True),
                    }
                if include_revrevs or only_revrevs:
                    possible_transpositions |= {
                        transposition(
                            framework,
                            (s, m),
                            (m, e),
                            inv_1=True,
                            inv_2=True,
                            revrev=True,
                        )
                    }
                rearrangements = rearrangements.union(possible_transpositions)

    if canonical_reps_only:
        rearrangements = __representatives(
            framework, rearrangements, prioritise_string_length=True
        )

    return rearrangements


def c_perm(n):
    """Return the permutation (1,...,n)(-n,...,1))"""
    pmN = structures.set_plus_minus(n)
    G = SymmetricGroup(pmN)
    c_string = f'({",".join(str(i) for i in list(range(1, n+1)))})({",".join(str(i) for i in list(range(-n, 0)))})'
    c = G(c_string)
    return c


def cut_positions(framework, sigma):
    """
    The set of cuts required to perform a rearrangement represented by sigma

    Args:
        framework (PositionParadigmFramework): The framework to create the transposition in
        sigma (Permutation): The permutation representing the rearrangement

    Returns:
        set: The set of cuts required to perform the rearrangement represented by sigma
    """
    sigma = deepcopy(framework.one_row(sigma))
    n = framework.n
    c = c_perm(n)
    sigma = list(sigma)
    return set(
        [
            tuple(([i + 1, c(i + 1)]))
            for i in range(len(sigma))
            if (sigma[c(i + 1) - 1] != c(sigma[i]))
        ]
    )


def __all_canonical_inversions(framework, num_regions=None):
    if not framework.oriented:
        raise NotImplementedError(f"not yet implemented for {str(framework)}")
    if framework.symmetry not in {SYMMETRY.circular, SYMMETRY.linear}:
        raise NotImplementedError(
            f"not implemented for framework with {str(framework.symmetry)} symmetry"
        )
    n = framework.n
    G = framework.genome_group()
    if num_regions == 1:
        return {G(f"({i},-{i})") for i in range(1, n + 1)} if framework.oriented else {}
    elif num_regions == 2:
        if framework.oriented:
            perms = {G(f"({i},{-1*(i+1)})({-1*i},{i+1})") for i in range(1, n)}
            if framework.symmetry is not SYMMETRY.linear:
                perms.union({G(f"({n},-1)(-{n},1)")})
        else:
            perms = {G(f"({i},{i+1})") for i in range(1, n)}
            if framework.symmetry is not SYMMETRY.linear:
                perms.union({G(f"(1,{n})")})
        return perms
    elif num_regions is None:  # Return all inversions up to length floor(n/2)
        up_to_length = floor(framework.n / 2) if framework.symmetry is SYMMETRY.circular else n
        if framework.oriented:
            perms = set()
            for permutation in G:
                cycle_type = list(permutation.cycle_type())
                if (
                    set(cycle_type) == {1, 2}
                    and list(cycle_type).count(2) <= up_to_length
                    and len(cut_positions(framework, permutation)) == 2
                ):
                    perms.add(permutation)
            if framework.symmetry is SYMMETRY.linear:
                perms = {
                    perm
                    for perm in perms
                    if not ("1" in str(perm) and str(n) in str(perm))
                }
            return perms
        else:
            raise NotImplementedError(f"not yet implemented for {framework}")
    else:
        raise NotImplementedError(
            f"inversions of length {num_regions} not yet implemented"
        )
    
_all_canonical_inversions = __all_canonical_inversions

def fast_all_inversion_reps(framework):
    n = framework.n
    G = framework.genome_group()
    n = int(floor(n / 2))
    elements = set()
    for i in range(1, n + 1):
        elt = ""
        for j in range(1, i + 1):
            elt += f"({j},-{i - j + 1})"
        elements.add(G(elt))
    return elements

def __two_region_adjacent_transposition_reps(framework):
    if not framework.oriented or framework.symmetry != SYMMETRY.circular:
        raise NotImplementedError(f"not yet implemented for {str(framework)}")
    G = framework.genome_group()
    return {G("(-2,-1)(1,2)"), G("(-2,1,2,-1)"), G("(-2,-1,2,1)"), G("(-2,2)(-1,1)")}


def double_coset(framework, perm):
    """Return the double coset of perm in the symmetry group of framework"""
    Z = framework.symmetry_group()
    return {d1 * perm * d2 for d1 in Z for d2 in Z}


def conjugacy_class(framework, perm):
    """Return the conjugacy class of perm in the symmetry group of framework"""
    Z = framework.symmetry_group()
    return {d.inverse() * perm * d for d in Z}


def single_coset(framework, perm):
    """Return the single coset of perm in the symmetry group of framework"""
    Z = framework.symmetry_group()
    return {perm * d for d in Z}


def segment_length(n, start, end):
    """Return the length of the segment from start to end"""
    return (end - start - 1) % n + 1


def segment_midpoint(n, start, end):
    """Return the midpoint of the segment from start to end"""
    return ((start + floor((segment_length(n, start, end) - 1) / 2) - 1) % n) + 1


def __representatives(
    framework,
    set_of_permutations,
    classes=CLASSES.double_cosets,
    prioritise_string_length=False,
):
    if not framework.oriented:
        raise NotImplementedError(f"not yet implemented for {str(framework)}")
    Z = framework.symmetry_group()
    perms = {x for x in set_of_permutations}
    cosets = []
    while len(perms):
        perm = perms.pop()
        if classes is CLASSES.double_cosets:
            coset = double_coset(framework, perm)
        elif classes is CLASSES.conjugacy_classes:
            coset = conjugacy_class(framework, perm)
        elif classes is CLASSES.cosets:
            coset = single_coset(framework, perm)
        for element in coset:
            perms.discard(element)
        cosets.append(coset)
    reps = set()

    if prioritise_string_length:
        sort_key = lambda x: (len(str(x).replace("-", "")), str(x))

    else:
        sort_key = lambda x: (
            sum(set(x.cycle_type()) - {1}),
            list(x.cycle_type()).count(2),
            str(x),
        )

    for coset in cosets:
        reps.add(
            sorted(
                list(coset),
                key=sort_key,
            )[0]
        )
    return reps


representatives = __representatives


def all_inversions_representatives(framework, num_regions=None):
    """
    Return the representatives of all inversions in the symmetry group of framework

    Args:
        framework (PositionParadigmFramework): The framework to create the transposition in
        num_regions (int, optional): The number of regions to invert. Defaults to None (all inversions).

    Returns:
        set: The set of representatives of all inversions in the symmetry group of framework
    """
    return __representatives(
        framework,
        __all_canonical_inversions(framework, num_regions=num_regions),
        classes=CLASSES.double_cosets,
    )

def all_adjacent_transpositions_representatives(framework, num_regions=None):
    if num_regions == 2:
        return __two_region_adjacent_transposition_reps(framework)
    else:
        raise NotImplementedError("model not yet implemented")


def permutation_with_cuts(framework, cuts, perm=None, start=None):
    """
    A generator for all permutations with the specified set of cuts.

    Args:
        framework (PositionParadigmFramework): The framework to create the transposition in
        cuts (set): The set of cuts to use

    Yields:
        Permutation: A permutation with the specified set of cuts
    """
    n = framework.n
    c = c_perm(n)
    if perm is None:
        if not (framework.oriented and framework.symmetry == SYMMETRY.circular):
            raise NotImplementedError(f"not yet implemented for {str(framework)}")
        perm = {}
        perm[1] = 1  # making the canonical instance
        results = permutation_with_cuts(cuts, perm, 2)
        for result in results:
            if result:
                yield list(result.values())
    else:  # recurse
        possibilities = set(range(1, n + 1)) | set(range(-n, 0))
        for val in perm.values():
            possibilities.remove(val)
            possibilities.remove(-val)
        if start - 1 in cuts:
            if c(perm[start - 1]) in possibilities:
                possibilities.remove(c(perm[start - 1]))
        else:
            possibilities = possibilities & {c(perm[start - 1])}
        if start == n:  # also check if *start* is a cut
            if start in cuts:
                # add exclusions
                if start in possibilities:
                    possibilities.remove(start)  # p[1]=1 so we can't have p[n]=n
            else:
                possibilities = possibilities & {n}  # we must choose n
            for possibility in possibilities:
                bperm = perm.copy()
                bperm[start] = possibility
                yield bperm
        else:
            for possibility in possibilities:
                bperm = perm.copy()
                bperm[start] = possibility
                for result in permutation_with_cuts(cuts, bperm, start + 1):
                    yield result



"""
A place for functions which return important but general algebraic structures requiring only a few parameters
"""

from sage.all_cmdline import SymmetricGroup


def set_plus_minus(n):
    return tuple(range(1, n + 1)) + tuple(range(-n, 0))


def HyperoctahedralGroup(n):
    """Return the hyperoctahedral group of order (2^n)n! as a subgroup of SymmetricGroup(2n) with two generators."""
    s_pmn = SymmetricGroup(set_plus_minus(n))
    if n == 1:
        return s_pmn
    h_gens = [
        s_pmn([(1, 2), (-1, -2)]),
        s_pmn(tuple(range(1, n + 1)) + tuple(range(-1, -(n + 1), -1))),
    ]
    return s_pmn.subgroup(h_gens)



"""
This file contains the class for the hyperoctahedral group.
"""

# %%
from cgt import structures
import sage
import numpy as np
from sage.all_cmdline import (
    SymmetricGroup,
    cartesian_product,
    GroupSemidirectProduct,
    SymmetricGroupRepresentation,
    SignedPermutations,
    matrix,
    block_matrix,
    Partitions,
    Permutation,
)


def HyperoctahedralGroupRepresentations(n):
    """
    This function returns the irreducible representations of the hyperoctahedral group of order 2^n * n!, indexed by pairs of partitions of l and m, where l + m = n.

    Args:
        n (int): The order of the hyperoctahedral group.

    Returns:
        dict: A dictionary of the irreducible representations of the hyperoctahedral group of order 2^n * n!
    """
    irreps = {}
    Hn = HyperoctahedralGroup(n)
    for little_subgroup_pair in Hn.little_subgroup_pairs():
        for partition_left in Partitions(little_subgroup_pair[0]):
            for partition_right in Partitions(little_subgroup_pair[1]):
                partition_pair = (partition_left, partition_right)
                irrep = Hn.irrep(partition_pair)

                def wrapped_irrep(elt, _irrep=irrep):
                    return _irrep(Hn.Phi(elt))

                irreps[partition_pair] = wrapped_irrep
    return irreps


class HyperoctahedralGroup:
    """
    This class represents the hyperoctahedral group of order 2^n * n!. It stores the group as a number of different isomorphic structures, as well as a number of other things that are helpful to keep track of in the context of genome rearrangements. It implements functions to compute the irreducible representations of the hyperoctahedral group.
    """

    def __init__(self, n, default_element_type="signed_permutation"):
        """
        Args:
            n (int): The order of the hyperoctahedral group.
            default_element_type (str, optional): The type of element to use when generating random elements. Defaults to "signed_permutation".
        """
        self._n = n
        self._signed_perms = structures.HyperoctahedralGroup(self._n)
        self._sage_signed_perms = SignedPermutations(self._n)
        self._Sn = SymmetricGroup(self._n)
        self._C2 = SymmetricGroup(2)
        self._N = self._signed_perms.subgroup(
            [self._signed_perms((i, -i)) for i in range(1, self._n + 1)]
        )
        self._Sn_in_Hn = self._signed_perms.subgroup(
            [
                self._signed_perms(
                    f'({",".join(str(x) for x in range(1,n+1))})({",".join(str(-x) for x in range(1,n+1))})'
                ),
                self._signed_perms(f"(1,2)(-1,-2)"),
            ]
        )
        self._Sn_wr_N = GroupSemidirectProduct(self._Sn_in_Hn, self._N)
        self._C2n = cartesian_product(tuple(SymmetricGroup(2) for _ in range(self._n)))
        self._semidirect_product = self.semidirect_product_with_hyperoctahedral_twist(
            self._Sn, self._C2n
        )

    def hyperoctahedral_twist(self, s, c):
        """
        This function implements the twist of the hyperoctahedral group, which is the action of S_n on (C_2)^n by permuting the factors.

        Args:
            s (Permutation): A permutation in S_n.
            c (tuple): A tuple of elements of C_2.

        Returns:
            an element of (C_2)^n: The tuple of elements of C_2 obtained by applying s to the tuple c.
        """
        try:
            _ = len(s)  # works if it is a "tuple"
        except:
            pass  # We have the usual hyperoctahedral group
        else:  # We have a wreath product of a little subgroup (as a direct product)
            #s = self._Sn(s[0]) * self._Sn(s[1])
            raise ValueError("This is not implemented yet")
        return self._C2n((c[s(k) - 1] for k in range(1, self._n + 1)))

    def orbit_representative(self, l, m):
        return self._C2n(tuple(() for _ in range(l)) + tuple((1, 2) for _ in range(m)))

    def little_subgroup_pairs(self):
        return tuple((l, self._n - l) for l in range(self._n + 1))

    def little_subgroup(self, l, m, as_direct_product=False):
        if l + m != self._n or l < 0 or m < 0 or self._n < 0:
            raise ValueError("l,m must be non-negative and l+m must equal n")
        if l == 0 or m == 0 and not as_direct_product:
            return self._Sn
        Sl = SymmetricGroup(list(range(1, l + 1)))
        if as_direct_product:
            Sm = SymmetricGroup(list(range(1, self._n - l + 1)))
            return cartesian_product((Sl, Sm))
        Sm = SymmetricGroup(list(range(l + 1, self._n + 1)))
        return self._Sn.subgroup(Sl.gens() + Sm.gens())

    def little_subgroup_pairs(self):
        return [(l, self._n - l) for l in range(self._n + 1)]

    def little_subgroups(self):
        return [
            self.little_subgroup(l, m) for l, m in self.little_subgroup_pairs(self._n)
        ]

    def orbit_representative(self, l, m):
        return self._C2n(tuple(() for _ in range(l)) + tuple((1, 2) for _ in range(m)))

    def character_for_tuple(self, c):
        return tuple(
            (lambda x: 1) if entry == self._C2(()) else (lambda x: x.sign())
            for entry in c
        )

    def representation_little_subgroup(self, partition_pair):
        rep_left = SymmetricGroupRepresentation(partition_pair[0])
        rep_right = SymmetricGroupRepresentation(partition_pair[1])
        l, m = sum([0] + list(partition_pair[0])), sum([0] + list(partition_pair[1]))
        n = l + m
        left_group = SymmetricGroup(l)
        right_group = SymmetricGroup(m)
        c = self.orbit_representative(l, m)
        x = self.character_for_tuple(c)

        def rep(g):
            try:
                _ = len(g[0])  # works if it is a "tuple"
            except:  # g[0] is not a "tuple" of permutations
                g = (self.element_in_little_subgroup(g[0], l, m), g[1])
            prod = rep_left(g[0][0]).tensor_product(rep_right(g[0][1]))
            character = (x[i](ci) for i, ci in enumerate(g[1]))
            coeff = 1
            for char in character:
                coeff *= char
            return coeff * prod

        return rep


    def _elt_in_subgroup(self, elt, l):
        return all(
            (
                all(int(x) <= l for x in str(cycle).strip("()").split(","))
                or all(int(x) > l for x in str(cycle).strip("()").split(","))
            )
            for cycle in elt
        )

    def irrep(self, partition_pair):
        """
        Returns the irreducible representation of the hyperoctahedral group of order 2^n * n! indexed by the pair of partitions of l and m where l + m = n. The representation is given as a function that takes an element of the hyperoctahedral group as input and returns the corresponding matrix.

        Args:
            partition_pair (tuple): A pair of partitions of l and m where l + m = n.

        Returns:
            function (group element -> real matrix): A function that takes an element of the hyperoctahedral group as input and returns the corresponding matrix.
        """
        rep = self.representation_little_subgroup(partition_pair)
        l, m = sum([0] + list(partition_pair[0])), sum([0] + list(partition_pair[1]))
        subgroup = self.little_subgroup(l, m)
        C2n_wr_SlxSm = self.semidirect_product_with_hyperoctahedral_twist(
            subgroup, self._C2n
        )
        tv = self.right_transversal_of_little_subgroup(l)
        id_rep = rep(C2n_wr_SlxSm.one())
        rep_dimension = (id_rep.nrows(), id_rep.ncols())

        def _irrep(g):
            Y = [[None for _ in range(len(tv))] for _ in range(len(tv))]
            for row in range(len(tv)):
                element_part = g * tv[row].inverse()
                for col in range(len(tv)):
                    element = tv[col] * element_part
                    if self._elt_in_subgroup(element[0], l):
                        Y[row][col] = np.array(rep(element), dtype=object)
                    else:  # A matrix of zeros
                        Y[row][col] = np.zeros(rep_dimension, dtype=sage.rings.rational.Rational)

            return matrix(np.block(Y))

        return _irrep

    def element_in_little_subgroup(self, elt, l, m):
        """
        Takes elt as an element of the little subgroup of Sn of order l!m! and returns an element of the direct product of S_l and S_m.
        """
        Sl = SymmetricGroup(l)
        Sm = SymmetricGroup(m)
        SlxSm = cartesian_product((Sl, Sm))
        cycles = list(elt)
        left = Sl.one()
        right = Sm.one()
        for cycle in cycles:
            if cycle in Sl:
                left *= Sl(cycle)
            else:
                string = str(cycle)
                for new, old in enumerate(range(l + 1, self._n + 1)):
                    string = string.replace(str(old), str(new + 1))
                right *= Sm(string)
        return SlxSm((left, right))

    def semidirect_product_with_hyperoctahedral_twist(self, G, N):
        """
        Compute a semi-direct product of H and G with the hyperoctahedral twist (G acts on H^n by permuting factors). The group returned is a group of pairs (c, s) where c is an element of H^n and s is an element of G. The group operation is given by (c1, s1) * (c2, s2) = (c1 * s1(c2), s1 * s2).

        Args:
            H (group): the direct product of n copies of some group.
            G (group): A group.

        Returns:
            group: The wreath product of H and G with the hyperoctahedral twist.
        """
        semidirect_product = GroupSemidirectProduct(
            G, N, act_to_right=True, twist=lambda s, c: self.hyperoctahedral_twist(s, c)
        )

        def random_element():
            Sn, C = semidirect_product.cartesian_factors()
            return semidirect_product((Sn.random_element(), C.random_element()))

        def order():
            return G.order() * N.order()

        def elements():
            return [semidirect_product((s, c)) for c in N for s in G]

        semidirect_product.random_element = random_element
        semidirect_product.action = self.hyperoctahedral_twist
        semidirect_product.order = order
        semidirect_product.elements = elements
        return semidirect_product

    def right_transversal(self, group, subgroup):
        transversal = [subgroup.one()]
        num_cosets = group.order() / subgroup.order()
        while len(transversal) < num_cosets:
            elt = group.random_element()
            if all(t * elt.inverse() not in subgroup for t in transversal):
                transversal.append(elt)
        return transversal

    def right_transversal_of_little_subgroup(self, k):
        from itertools import combinations

        n = self._n
        H = self
        combs = list(combinations(tuple((i for i in range(1, n + 1))), k))
        t_new = []
        for comb in combs:
            sets = ((i for i in range(1, k + 1)), (i for i in range(k + 1, n + 1)))
            elt = Permutation(
                [next(sets[0]) if i in comb else next(sets[1]) for i in range(1, n + 1)]
            )
            t_new.append(
                H._semidirect_product(
                    (H._Sn(elt.cycle_string()), (() for _ in range(n)))
                ).inverse()
            )
        return t_new

    def Phi(self, elt):
        c = tuple(self._C2(() if elt.inverse()(k) > 0 else (1, 2)) for k in range(1, self._n + 1))
        s = self._Sn(Permutation([abs(elt(k)) for k in range(1, self._n + 1)]))
        return self._semidirect_product((s, c))


# %%



"""
Implements a number of distance measures for genomes under the position paradigm.

The primary function of this module is the distance_matrix function, which returns a distance matrix for a given set of genomes and distance measure.

The distance measures implemented are:
    - min: the minimum number of rearrangements required to transform one genome into another
    - min_weighted: the minimum number of rearrangements required to transform one genome into another, weighted by the inverse of the probability of the rearrangement
    - MFPT: the mean first passage time from the identity to a given genome, where the target is an absorbing state
    - MLE: the maximum likelihood estimate of the time elapsed between the identity and a given genome

Individual likelihood functions for the time elapsed between the identity and a given genome can also be obtained
"""

from cgt.enums import ALGEBRA, DISTANCE, DATA, SYMMETRY
from cgt.constants import VALUES_OF_N_WITH_SAVED_IRREPS_OF_Z
import numpy as np
import networkx as nx
from sage.all import matrix, real, exp, round
import scipy
from scipy.optimize import minimize_scalar
from cgt import pickle_manager
from warnings import warn


def mles(framework, model, genome_instances=None, verbose=False, show_work=False):
    """
    Returns a dictionary of maximum likelihood estimates for each genome instance under the given model and framework.

    Args:
        framework (PositionParadigmFramework): the framework
        model (Model): the model
        genome_instances (list): a list of genome instances to compute MLEs for. If None, all genomes in the framework are used.
        verbose (bool): whether to print progress

    Returns:
        dict: a dictionary of maximum likelihood estimates, indexed by genome instance
    """
    mles = {}
    if genome_instances is None:
        genome_instances = [
            framework.canonical_instance(g) for g in framework.genomes()
        ]
    for instance in genome_instances:
        if verbose:
            print(f"Computing MLE for {instance}")
        mles[instance] = mle(framework, model, instance, show_work=show_work)
    return mles


def distance(framework, model, genome_instance, distance_measure):
    distance_func = get_distance_function(distance_measure)
    return distance_func(framework, model, [genome_instance])


def distance_between(framework, model, genome_instance_1, genome_instance_2, distance_measure):
    return distance(framework, model, genome_instance_1.inverse() * genome_instance_2, distance_measure)


def mle(framework, model, genome_instance, show_work=False):
    """
    Returns the maximum likelihood estimate for a given genome instance under the given model and framework.

    Args:
        framework (PositionParadigmFramework): the framework
        model (Model): the model
        genome_instance (group element): the genome instance to compute the MLE for

    Returns:
        float: the maximum likelihood estimate
    """
    L = likelihood_function(framework, model, genome_instance)
    max_t = maximise(framework, L)
    if show_work:
        return max_t, L
    else:
        return max_t


def maximise(framework, L, max_time=None):
    """
    Return the time that maximises likelihood function L, using additional information from the framework

    Args:
        framework (PositionParadigmFramework): the framework
        L (function): the likelihood function to maximise
        max_time (float): the maximum time to consider (default: 100)

    Returns:
        float: the time that maximises the likelihood function
    """
    if max_time is None: 
        max_time = framework.n * 25
    limit = 1 / framework.num_genomes()
    t_max = minimize_scalar(
        lambda t: -1 * L(t), method="bounded", bounds=(0, max_time)
    )["x"]
    mle = t_max if L(t_max) > limit * (1 + (1/3)*0.01) else np.nan

    return mle


def _projection_operators(mat, eigs):
    """
    Return projection operators for given matrix and its eigenvalues

    Args:
        mat (matrix): a representation of zs
        eigs (list): a list of eigenvalues of mat

    Returns:
        list: a list of projection operators
    """
    dim = mat.nrows()
    projections = [matrix.identity(dim) for _ in eigs]
    for e1, eig1 in enumerate(eigs):
        for eig2 in eigs:
            if eig1 != eig2:
                projections[e1] *= (mat - (eig2 * matrix.identity(dim))) * (
                    1 / (eig1 - eig2)
                )
    return projections


def _irreps_of_zs(framework, model, force_recompute=False):
    """
    Return a set of matrices---images of zs under each irrep

    Args:
        framework (PositionParadigmFramework): the framework
        model (Model): the model
        force_recompute (bool): whether to force recomputation and invalidate cache (default: False)

    Returns:
        list: a list of irredicuble representations of the group applied to zs
    """
    key = DATA.irreps_zs
    if key in model.data_bundle and not force_recompute:
        irreps_of_zs = model.data_bundle[key]
    else:
        z = framework.symmetry_element()
        s = model.s_element(in_algebra=ALGEBRA.genome)
        irreps_of_z = _irreps_of_z(framework, model)
        irreps_of_s = framework.irreps(s)
        irreps_of_zs = [
            irrep_z * irrep_s for irrep_z, irrep_s in zip(irreps_of_z, irreps_of_s)
        ]
        model.data_bundle[key] = irreps_of_zs
    return irreps_of_zs


def _irreps_of_z(framework, model=None, force_recompute=False):
    """
    Return a set of matrices---images of z (the symmetry element) under each irrep

    Args:
        framework (PositionParadigmFramework): the framework
        model (Model): the model
        force_recompute (bool): whether to force recomputation and invalidate cache (default: False)

    Returns:
        list: a list of irredicuble representations of the group applied to z
    """
    key = DATA.irreps_z
    if model is not None and key in model.data_bundle and not force_recompute:
        return model.data_bundle[key]
    if framework.n in VALUES_OF_N_WITH_SAVED_IRREPS_OF_Z and framework.symmetry is SYMMETRY.circular and not force_recompute:
        try:
            irreps_of_z = pickle_manager.retrieve_irrep_of_z(n=framework.n)
            if model is not None: 
                model.data_bundle[key] = irreps_of_z
                model.data_bundle[DATA.irreps_z_np] = [irrep.numpy() for irrep in irreps_of_z] # TODO: This is a hack
            return irreps_of_z
        except:
            pass
    z = framework.symmetry_element()
    irreps_of_z = framework.irreps(z)
    if model is not None: 
        model.data_bundle[key] = irreps_of_z
        model.data_bundle[DATA.irreps_z_np] = [irrep.numpy() for irrep in irreps_of_z] # TODO: This is a hack
    return irreps_of_z


def _eigenvalues(mat, round_to=7, make_real=True, inc_repeated=False):
    """Return all the eigenvalues for a given matrix mat"""
    col = list if inc_repeated else set
    all_eigs = np.linalg.eigvals(mat)
    return sorted(
        col(round(real(eig) if make_real else eig, round_to) for eig in all_eigs)
    )


def _eigen_data(framework, model, irreps_of_zs, round_vecs_to=10, round_vals_to=7):
    """
    Return an eigenvector matrix for each irrep of zs. Also return eigenvalue and eigenvector lists

    Args:
        irreps_of_zs (list): a list of irreps of zs
        round_vecs_to (int): the number of decimal places to round eigenvectors to (default: 10)
        round_vals_to (int): the number of decimal places to round eigenvalues to (default: 7)

    Returns:
        list: a list of eigenvector matrices
        list: a list of eigenvalue lists
        list: a list of eigenvector lists
    """
    eigen_data = {
        DATA.eigval_lists: [],
        DATA.eigval_sets: [],
        DATA.eigvec_lists: [],
        DATA.eigvec_mat_inv: [],
    }

    # Fix for weird issue with SVD convergence (n=9)
    if framework.n == 9 and round_vecs_to == 10:
        round_vecs_to = 12

    # If the data is already stored in the model, return it
    if DATA.eig_data in model.data_bundle:
        return model.data_bundle[DATA.eig_data]

    # Otherwise compute the eigen_data dictionary
    for irrep_of_zs in irreps_of_zs:
        # Prepare irrep matrices
        irrep_zs_np = irrep_of_zs.numpy()
        # Compute eigenvalues and eigenvectors
        eigenvalues, eigenvectors = np.linalg.eig(irrep_zs_np)
        # Rounding
        eigenvectors = np.around(eigenvectors, round_vecs_to)  # a matrix
        #print(f"Condition number is: {np.linalg.cond(eigenvectors)}")
        eigenvalues = [round(val, round_vals_to) for val in eigenvalues]
        # Store data
        eigen_data[DATA.eigval_lists].append(eigenvalues)
        eigen_data[DATA.eigval_sets].append(set(eigenvalues))
        eigen_data[DATA.eigvec_lists].append(eigenvectors.T)
        inverse_of_eigenvector_mat = np.linalg.pinv(eigenvectors)
        #print(f"Elements of inverted eigenvector matrix range from ({np.amin(inverse_of_eigenvector_mat)} to {np.amax(inverse_of_eigenvector_mat)})")
        eigen_data[DATA.eigvec_mat_inv].append(inverse_of_eigenvector_mat)


    # Store the data in the model
    model.data_bundle[DATA.eig_data] = eigen_data
    return eigen_data


def probability_to_reach_in_steps(framework, model, sigma, k):
    alpha = prob_to_reach_in_steps_func(framework, model, sigma)
    return alpha(k)


def prob_to_reach_in_steps_func(framework, model, sigma, use_eigenvectors=True):
    G, Z = framework.genome_group(), framework.symmetry_group()
    instance = framework.cycles(sigma)
    irreps = framework.irreps()
    irreps_of_zs = _irreps_of_zs(framework, model)
    irreps_of_z = _irreps_of_z(framework, model)
    if DATA.eigval_sets_old in model.data_bundle:
        eig_lists = model.data_bundle[DATA.eigval_sets_old]
    else:
        eig_lists = [_eigenvalues(irrep_zs) for irrep_zs in irreps_of_zs]
        model.data_bundle[DATA.eigval_sets_old] = eig_lists
    if not use_eigenvectors:
        if DATA.projections in model.data_bundle:
            projections = model.data_bundle[DATA.projections]
        else:
            projections = [
                _projection_operators(*vals) for vals in zip(irreps_of_zs, eig_lists)
            ]
            model.data_bundle[DATA.projections] = projections
        traces = _partial_traces_for_genome(
            framework, model, instance, irreps, irreps_of_zs, projections, eig_lists, irreps_of_z
        )
    else:
        eig_lists, traces = _partial_traces_for_genome_using_eigenvectors(
            framework, model, instance, irreps, irreps_of_zs, irreps_of_z
        )
    dims = [irrep_of_zs.nrows() for irrep_of_zs in irreps_of_zs]

    def prob_in_steps(k):
        alpha = 0
        for ptrace, dim, eig_list in zip(traces.values(), dims, eig_lists):
            term = 0
            for e, eig in enumerate(eig_list):
                term += (eig**k) * ptrace[eig]
            alpha += dim * term
        return (Z.order() / G.order()) * alpha

    return prob_in_steps


def discrete_MFPT(framework, model, genome_reps=None, verbose=False):
    if genome_reps is None:
        genomes = list(framework.genomes().keys())
        genome_reps = genomes
    dists = {}
    a_star = framework.symmetry_group().order() / framework.genome_group().order()
    identity = framework.one_row(framework.genome_group().identity())
    for instance in genome_reps:
        if instance == identity:
            dists[instance] = 0
        else:
            a = prob_to_reach_in_steps_func(framework, model, instance)
            summands = [0]
            a_values = [0]
            k = 1
            prod = 1
            while True:
                a_k = a(k)
                a_values.append(a_k)
                prod *= 1 - a_values[k - 1]
                summands.append(k * a_k * prod)
                if (abs(a_values[k] - a_values[k - 1]) < 10 ** (-3)) and (
                    abs(a_values[k] - a_star) < 10 ** (-3)
                ):
                    if verbose:
                        print(
                            f"Terms {a_values[k]} and {a_values[k-1]} are close enough to {a_star}"
                        )
                    break
                k += 1
            c = k
            if verbose:
                print(f"Converged at {k=}")
            adjustment_term = sum(
                j * a_star * (1 - a_star) ** j for j in range(0, c + 1)
            )
            remaining_sum = ((prod * a_star) / ((1 - a_star) ** (c + 1))) * (
                ((1 - a_star) / (a_star**2)) - adjustment_term
            )
            dists[instance] = sum(summands) + remaining_sum
    return dists


def _partial_traces_for_genome(
    framework,
    model,
    instance,
    irreps,
    irreps_of_zs,
    projections,
    eig_lists,
    irreps_of_z=None,
):
    """Return dictionary of partial traces, indexed first by irrep index and then by eigenvalaue"""
    instance_inverse = instance.inverse()
    if (
        irreps_of_z is None
    ):  # It almost never is. Can't used the cached version below because it's stored in the model (which we don't have)
        irreps_of_z = framework.irreps(framework.symmetry_element())
    traces = {
        r: {eigenvalue: {} for eigenvalue in eig_lists[r]}
        for r in range(len(irreps_of_zs))
    }
    for r, irrep in enumerate(irreps):  # Iterate over irreducible representations
        zero_irrep = not irreps_of_z[r] 
        if zero_irrep:  # matrix of zeros
            sigd = irreps_of_z[r]
        else:
            sigd = irreps_of_z[r] * irrep(instance_inverse)
        for e, eigenvalue in enumerate(eig_lists[r]):
            if zero_irrep:  # matrix of zeros
                traces[r][eigenvalue] = 0
            else:
                traces[r][eigenvalue] = real((sigd.numpy() @ projections[r][e]).trace())
    return traces


def _partial_traces_for_genome_using_eigenvectors(
    framework, model, instance, irreps, irreps_of_zs, irreps_of_z
):
    """Return dictionary of partial traces, indexed first by irrep index and then by eigenvalaue"""
    instance_inverse = instance.inverse()
    eigen_data = _eigen_data(framework, model, irreps_of_zs)
    if DATA.partial_traces in model.data_bundle and instance_inverse in model.data_bundle[DATA.partial_traces]:
        return eigen_data[DATA.eigval_sets], model.data_bundle[DATA.partial_traces][instance_inverse]
    traces = {}
    irreps_of_z_np = model.data_bundle[DATA.irreps_z_np]
    for r, irrep in enumerate(irreps):  # Iterate over irreducible representations
        # Faster convert to numpy
        traces[r] = {}
        irrep_instance = irrep(instance_inverse)
        irrep_instance_np = np.zeros(irrep_instance.dimensions())
        for (i, j), val in irrep_instance.items():
            irrep_instance_np[i,j] = val
        eigenvalue_list = eigen_data[DATA.eigval_lists][r]
        if not irreps_of_z[r]:  # matrix of zeros
            irrep_of_g_inverse_z_np = irreps_of_z[r]
            traces[r] = {eigenvalue: 0 for eigenvalue in eigenvalue_list}
        else:
            irrep_of_g_inverse_z_np = irreps_of_z_np[r] @ irrep_instance_np
            eigenvectors = eigen_data[DATA.eigvec_lists][r]
            P_inv = eigen_data[DATA.eigvec_mat_inv][r]
            mat = ((P_inv @ irrep_of_g_inverse_z_np) * eigenvectors).sum(-1)
            for v, eig in enumerate(eigenvalue_list):
                if eig in traces[r]:
                    traces[r][eig] += mat[v]
                else:
                    traces[r][eig] = mat[v]
        full_trace = irrep_of_g_inverse_z_np.trace()
        sum_of_partial_traces = sum(traces[r].values())
        if not sum_of_partial_traces - full_trace < 1e-04:
            print(f"Sum of partial traces ({round(sum_of_partial_traces, 5)}) is very different from the full trace ({round(full_trace, 5)}) for irrep {r}.")
    
    if DATA.partial_traces not in model.data_bundle:
        model.data_bundle[DATA.partial_traces] = {}
    model.data_bundle[DATA.partial_traces][instance_inverse] = traces

    return eigen_data[DATA.eigval_sets], traces


def _eig_lists(model, irreps_of_zs):
    if DATA.eigval_sets_old in model.data_bundle:
        eig_lists = model.data_bundle[DATA.eigval_sets_old]
    else:
        eig_lists = [_eigenvalues(irrep_zs) for irrep_zs in irreps_of_zs]
        model.data_bundle[DATA.eigval_sets_old] = eig_lists
    return eig_lists


def _projections(model, irreps_of_zs, eig_lists):
    if DATA.projections in model.data_bundle:
        projections = model.data_bundle[DATA.projections]
    else:
        projections = [
            _projection_operators(*vals) for vals in zip(irreps_of_zs, eig_lists)
        ]
        model.data_bundle[DATA.projections] = projections
    return projections


def likelihood_function(framework, model, genome, use_eigenvectors=True):
    """Return the likelihood function for a given genome"""
    instance = framework.cycles(genome)
    G, Z = framework.genome_group(), framework.symmetry_group()
    irreps = framework.irreps()
    irreps_of_zs = _irreps_of_zs(framework, model)
    irreps_of_z = _irreps_of_z(framework, model)
    if not use_eigenvectors:
        eig_lists = _eig_lists(model, irreps_of_zs)
        projections = _projections(model, irreps_of_zs, eig_lists)
        traces = _partial_traces_for_genome(
            framework,
            model,
            instance,
            irreps,
            irreps_of_zs,
            projections,
            eig_lists,
            irreps_of_z,
        )
    else:
        eig_lists, traces = _partial_traces_for_genome_using_eigenvectors(
            framework, model, instance, irreps, irreps_of_zs, irreps_of_z
        )
    dims = [irrep_of_zs.nrows() for irrep_of_zs in irreps_of_zs]

    def likelihood(t):
        ans = 0
        for r, dim in enumerate(dims):
                ans += (
                    Z.order()
                    * (exp(-t) / G.order())
                    * dim
                    * sum(
                        exp(eigenvalue * t) * traces[r][eigenvalue]
                        for eigenvalue in eig_lists[r]
                    )
                )
        return real(ans)

    return likelihood


def min_distance(framework, model, genome_reps=None, weighted=False):
    """Return dictionary of minimum distances ref->genome, using inverse of model probabilities as weights (or not)"""
    matrix = model.reg_rep_of_zs().toarray()
    if weighted:
        matrix[matrix != 0] = 1 / matrix[matrix != 0]
    graph = nx.Graph(matrix)
    genomes = list(framework.genomes().keys())
    if genome_reps is None:
        genome_reps = genomes
    return {
        rep: nx.shortest_path_length(
            graph,
            source=0,
            target=genomes.index(rep),
            weight="weight" if weighted else None,
        )
        for r, rep in enumerate(genome_reps)
    }


def min_distance_using_irreps(framework, model, genome_reps=None):
    num_genomes = int(
        framework.genome_group().order() / framework.symmetry_group().order()
    )
    if genome_reps is None:
        genomes = list(framework.genomes().keys())
        genome_reps = genomes
    dists = {}
    for rep in genome_reps:
        a = prob_to_reach_in_steps_func(framework, model, rep)
        for k in range(0, num_genomes):
            if a(k) > 10 ** (-7):
                dists[rep] = k
                break
    return dists

def first_nonzero_value(framework, function, limit=None):
    if limit is None:
        limit = framework.genome_group().order() / framework.symmetry_group().order()
    first_k = limit
    for k in range(1, limit):
        if function(k) > 10 ** (-8):
            first_k = k
            break
    return first_k

def MFPT(framework, model, genome_reps=None, scale_by=1):
    """Return the mean time elapsed rearranging ref->genome where the target is an absorbing state"""
    genomes = framework.genomes()
    reps = list(genomes.keys())
    P = model.reg_rep_of_zs().toarray()
    P_0 = P.copy() 
    for i in range(len(P)):
        P_0[0, i] = 0
    Z = np.linalg.pinv(np.eye(len(P)) - P_0)

    def MFTP_dists():
        return (Z @ P_0 @ Z)[:, 0]

    MFTP_distances = list(MFTP_dists())
    MFTP_distances = {rep: scale_by * MFTP_distances[r] for r, rep in enumerate(reps)}
    if genome_reps is not None:
        MFTP_distances = {
            rep: MFTP_distances[framework.canonical_instance(rep)]
            for rep in genome_reps
        }
    return MFTP_distances

def fast_MFPT(framework, model):
    reg_rep, genomes = framework.fast_reg_rep_of_zs(model)
    Q = reg_rep[1:,1:] # Remove the absorbing state
    m = Q.shape[0] # n - 1
    A = scipy.sparse.identity(m) - Q
    scipy.sparse.linalg.cg(A, np.ones(m))

def dict_to_distance_matrix(distances, framework, genomes=None):
    """If need to convert to pairwise distances, supply a list of genomes."""
    if genomes is not None:
        D = np.zeros((len(genomes), len(genomes)))
        for i in range(len(genomes)):
            canonical_i = framework.canonical_instance(
                framework.random_instance(genomes[i])
            )
            distances_copy = {canonical_i * k: v for k, v in distances.items()}
            for j in range(len(genomes)):
                canonical_j = framework.canonical_instance(
                    framework.random_instance(genomes[j])
                )
                D[i, j] = distances_copy[canonical_j]
    else:
        D = np.zeros((len(distances), len(distances)))
        for (i, j), distance in distances.items():
            D[i, j] = distance
    return D


def genomes_for_dist_matrix(framework, genomes):
    instances = [framework.canonical_instance(g) for g in genomes]
    # Get the genomes g that we need dist(id -> g) for:
    need_distances = {}
    for i, _ in enumerate(genomes):
        for j, _ in enumerate(genomes):
            if i < j:
                canonical_i, canonical_j = instances[i], instances[j]
                need_distances[(i, j)] = canonical_i.inverse() * canonical_j
    return need_distances


def get_distance_function(distance_type):
    match distance_type:
        case DISTANCE.MFPT: return MFPT
        case DISTANCE.discrete_MFPT: return discrete_MFPT
        case DISTANCE.min: return min_distance_using_irreps
        case DISTANCE.MLE: return mles
        case DISTANCE.min_weighted: return min_distance


def distance_matrix(
    framework,
    model,
    genomes,
    distance,
    replace_nan_with=np.nan
):
    """
    Compute a distance matrix for a given set of genomes and distance measure.

    Args:
        framework (PositionParadigmFramework): the framework
        model (Model): the model
        genomes (list): a list of genomes to compute distances for
        distance (DISTANCE): the distance measure to use
    """
    need_distances = genomes_for_dist_matrix(framework, genomes)
    pairs, need_distances = map(list, zip(*need_distances.items()))

    # Parameters
    params = { "framework" : framework, "model" : model, "genome_reps" : need_distances }

    # Compute the distances:
    distance_func = get_distance_function(distance)
    distances = distance_func(**params)

    # Construct the distance matrix
    D = np.zeros((len(genomes), len(genomes)))
    for p, (i, j) in enumerate(pairs):
        D[i, j] = distances[need_distances[p]]

    if replace_nan_with != np.nan:
        D[np.isnan(D)] = replace_nan_with

    return D
